/*
 * Copyright (c) 2016-2017,2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Aug 23 2016
 */
#include "aclibmrsim.h"

#include <math.h>
#include <stdlib.h>

#include <Eigen/Geometry>

#include "private.h"


/* --- mrsim_update_wrench ------------------------------------------------- */

void
mrsim_update_wrench(const struct mrsim_s *mrsim)
{
  unsigned int i, j;

  for(i = 0; i < 3; i++) {
    mrsim->wrench->thrust[i] = 0.;
    mrsim->wrench->torque[i] = 0.;
    for(j = 0; j < mrsim->params->body.rotors; j++) {
      mrsim->wrench->thrust[i] +=
        mrsim->params->body.G[i*mrsim->params->body.rotors + j] *
        mrsim->motor[j].state.w * mrsim->motor[j].state.w;
      mrsim->wrench->torque[i] +=
        mrsim->params->body.G[(i+3)*mrsim->params->body.rotors + j] *
        mrsim->motor[j].state.w * mrsim->motor[j].state.w;
    }
  }
}


/* --- mrsim_update_imu ---------------------------------------------------- */

void
mrsim_update_imu(const struct mrsim_s *mrsim)
{
  using namespace Eigen;

  const Vector3d g(0., 0., - mrsim->params->body.g);
  Quaternion<double> q(
    mrsim->body->qw, mrsim->body->qx, mrsim->body->qy, mrsim->body->qz);

  Vector3d w(mrsim->body->wx, mrsim->body->wy, mrsim->body->wz);
  Vector3d a(mrsim->body->ax, mrsim->body->ay, mrsim->body->az);

  Map<Vector3d> wb(mrsim->imu->w);
  Map<Vector3d> ab(mrsim->imu->a);

  wb <<
    mrsim_gnoise(mrsim->params->state.wstddev),
    mrsim_gnoise(mrsim->params->state.wstddev),
    mrsim_gnoise(mrsim->params->state.wstddev);
  wb += q.conjugate() * w;

  ab <<
    mrsim_gnoise(mrsim->params->state.astddev),
    mrsim_gnoise(mrsim->params->state.astddev),
    mrsim_gnoise(mrsim->params->state.astddev);
  ab += q.conjugate() * (a - g);
}


/* --- mrsim_update_body --------------------------------------------------- */

void
mrsim_update_body(const struct mrsim_s *mrsim)
{
  const Eigen::Matrix3d J(mrsim->params->body.J);
  const Eigen::Vector3d thrust(mrsim->wrench->thrust);
  const Eigen::Vector3d torque(mrsim->wrench->torque);
  const Eigen::Vector3d g(0, 0, - mrsim->params->body.g);
  const double dt = mrsim->ts->dt;

  /* current state */
  Eigen::Vector3d p(mrsim->body->px, mrsim->body->py, mrsim->body->pz);
  Eigen::Quaternion<double> q(
    mrsim->body->qw, mrsim->body->qx, mrsim->body->qy, mrsim->body->qz);
  Eigen::Vector3d v(mrsim->body->vx, mrsim->body->vy, mrsim->body->vz);
  Eigen::Vector3d w(mrsim->body->wx, mrsim->body->wy, mrsim->body->wz);
  Eigen::Vector3d a = g + q * thrust / mrsim->params->body.m;


  /* update position */
  p += v * dt + a * dt*dt/2;

  /* update linear velocity */
  v += a * dt;

  /* update angular velocity */
  Eigen::Vector3d wb, dwb;
  wb = q.conjugate() * w;
  dwb = J.inverse()*((J*wb).cross(wb) + torque);
  wb += dwb * dt;
  w = q * wb;

  /* update attitude */
  Eigen::Quaternion<double> dq;
  double a2 = dt * dt * w.squaredNorm();
  if (a2 < 1e-1) {
    dq.w() = 1 - a2/8 /*std::cos(a/2)*/;
    dq.vec() = (0.5 - a2/48 /*std::sin(a/2)/a*/) * dt * w;
  } else {
    double a = std::sqrt(a2);
    dq.w() = std::cos(a/2);
    dq.vec() = std::sin(a/2)/a * dt * w;
  }
  q = dq * q;


  /* update state */
  mrsim->body->px = p(0);
  mrsim->body->py = p(1);
  mrsim->body->pz = p(2);
  mrsim->body->qw = q.w();
  mrsim->body->qx = q.vec()(0);
  mrsim->body->qy = q.vec()(1);
  mrsim->body->qz = q.vec()(2);

  mrsim->body->vx = v(0);
  mrsim->body->vy = v(1);
  mrsim->body->vz = v(2);
  mrsim->body->wx = w(0);
  mrsim->body->wy = w(1);
  mrsim->body->wz = w(2);

  mrsim->body->ax = a(0);
  mrsim->body->ay = a(1);
  mrsim->body->az = a(2);
}


/* --- mrsim_gnoise -------------------------------------------------------- */

double
mrsim_gnoise(double stddev)
{
  double u1, u2;
  double s;

  do {
    u1 = 2 * drand48() - 1; /* [-1; 1] */
    u2 = 2 * drand48() - 1;

    s = u1*u1 + u2*u2;
  } while(s >= 1.);

  return sqrt(-2. * log(s)/s) * stddev * u1;
}
